"""
FastAPI Monitoring SDK 타입 정의
로그 및 분산 추적을 지원함 (OpenTelemetry 호환)
"""

from typing import TypedDict, Literal, Optional, Union
from dataclasses import dataclass


class MonitoringConfig(TypedDict, total=False):
    """SDK 설정 인터페이스"""

    # 필수 설정
    api_key: str
    """Producer 서버 인증을 위한 API 키"""

    service_name: str
    """서비스 이름 식별자"""

    # 선택 설정
    endpoint: Optional[str]
    """Producer 서버 엔드포인트 URL (예: https://producer.woongno-monitoring.com)"""

    log_endpoint: Optional[str]
    """Log 전송 엔드포인트 URL (예: https://producer.woongno-monitoring.com/sdk/logs)"""

    trace_endpoint: Optional[str]
    """Trace 전송 엔드포인트 URL (예: https://producer.woongno-monitoring.com/sdk/traces)"""

    environment: Optional[str]
    """환경 설정 (예: production, staging, development)"""

    batch_size: Optional[int]
    """데이터 전송을 위한 배치 크기 (기본값: 100)"""

    flush_interval: Optional[int]
    """데이터 플러시 간격 (초 단위, 기본값: 5)"""

    enable_log_tracking: Optional[bool]
    """로그 추적 활성화 여부 (기본값: True)"""

    enable_http_tracking: Optional[bool]
    """HTTP 요청 추적 활성화 여부 (기본값: True)"""

    enable_http_client_tracking: Optional[bool]
    """외부 HTTP 클라이언트 추적 활성화 여부 (기본값: True)"""

    enable_bedrock_tracking: Optional[bool]
    """Bedrock API 호출 추적 활성화 여부 (기본값: True)"""


class LogEntry(TypedDict, total=False):
    """로그 엔트리 인터페이스"""

    type: Literal['log']
    timestamp: str
    service_name: str
    environment: str
    level: str
    message: str
    context: Optional[str]
    trace: Optional[str]
    trace_id: Optional[str]
    """Trace ID - 트레이스 도중 발생한 로그인 경우 상속됨, 아니면 None"""


class SpanEntry(TypedDict, total=False):
    """분산 추적을 위한 Span 엔트리 인터페이스"""

    type: Literal['span']
    timestamp: str
    service_name: str
    environment: str

    trace_id: str
    """Trace ID (32자 hex 문자열) - 전체 요청 추적을 식별함"""

    span_id: str
    """Span ID (16자 hex 문자열) - 특정 span을 식별함"""

    parent_span_id: Optional[str]
    """Parent Span ID (root span의 경우 None)"""

    name: str
    """Span 이름 (예: "POST /payments", "SELECT users")"""

    kind: Literal['SERVER', 'CLIENT']
    """Span 종류"""

    duration_ms: float
    """소요 시간 (밀리초 단위)"""

    status: Literal['OK', 'ERROR']
    """상태"""

    # HTTP span 속성들
    http_method: Optional[str]
    http_path: Optional[str]
    http_url: Optional[str]
    http_status_code: Optional[int]

    # Database span 속성들
    db_system: Optional[str]
    db_statement: Optional[str]
    db_operation: Optional[str]

    # Bedrock span 속성들
    bedrock_model_id: Optional[str]
    bedrock_operation: Optional[str]
    bedrock_input_tokens: Optional[int]
    bedrock_output_tokens: Optional[int]


# 통합 데이터 엔트리 타입
DataEntry = Union[LogEntry, SpanEntry]


@dataclass
class TraceContext:
    """ContextVar에 저장되는 Trace Context 데이터클래스"""

    trace_id: str
    span_id: str
    parent_span_id: Optional[str] = None


@dataclass
class SpanOptions:
    """새 span 생성을 위한 옵션 데이터클래스"""

    name: str
    kind: Literal['SERVER', 'CLIENT']
    attributes: Optional[dict] = None


@dataclass
class ActiveSpan:
    """활성 Span 추적 데이터클래스"""

    trace_id: str
    span_id: str
    parent_span_id: Optional[str]
    name: str
    kind: Literal['SERVER', 'CLIENT']
    start_time: float
    attributes: dict



"""
Bedrock API 호출 수집기
boto3를 통한 Bedrock API 호출을 추적함

역할: boto3 bedrock-runtime 클라이언트의 API 호출을 자동으로 추적함
사용: MonitoringSDK.init()에서 자동 등록됨
"""

import time
import json
import io
from datetime import datetime
from typing import Any, Dict, Optional
from ..types.interfaces import SpanEntry
from ..transport.batch_sender import BatchSender
from ..tracing.trace_context import get_trace_context
from ..tracing.id_generator import generate_span_id


class BedrockCollector:
    """
    Bedrock API 호출을 추적하는 수집기
    """

    def __init__(self, batch_sender: BatchSender, service_name: str, environment: str):
        """
        BedrockCollector 초기화

        Args:
            batch_sender: 배치 전송자
            service_name: 서비스 이름
            environment: 환경 설정
        """
        self.batch_sender = batch_sender
        self.service_name = service_name
        self.environment = environment

    def setup_bedrock_instrumentation(self) -> None:
        """
        boto3 bedrock-runtime 클라이언트에 instrumentation을 설정함

        역할: bedrock-runtime API 호출 시 자동으로 추적하도록 설정
        """
        try:
            import boto3
            from botocore.client import BaseClient

            # boto3 client의 _make_api_call 메서드를 패치
            original_make_api_call = BaseClient._make_api_call

            def patched_make_api_call(client_self, operation_name, api_params):
                # bedrock-runtime 서비스만 추적
                service_name = client_self._service_model.service_name

                if service_name != "bedrock-runtime":
                    return original_make_api_call(
                        client_self, operation_name, api_params
                    )

                trace_context = get_trace_context()
                if not trace_context:
                    return original_make_api_call(
                        client_self, operation_name, api_params
                    )

                start_time = time.time()
                status = "ERROR"
                response = None
                error = None

                try:
                    response = original_make_api_call(
                        client_self, operation_name, api_params
                    )
                    status = "OK"
                    return response
                except Exception as e:
                    status = "ERROR"
                    error = e
                    raise
                finally:
                    duration = (time.time() - start_time) * 1000
                    self._record_bedrock_span(
                        operation_name,
                        api_params,
                        response,
                        duration,
                        status,
                        trace_context.trace_id,
                        trace_context.span_id,
                        error,
                    )

            BaseClient._make_api_call = patched_make_api_call

        except ImportError:
            # boto3가 설치되지 않은 경우 무시
            pass

    def _record_bedrock_span(
        self,
        operation_name: str,
        api_params: Dict[str, Any],
        response: Optional[Dict[str, Any]],
        duration: float,
        status: str,
        trace_id: str,
        parent_span_id: str,
        error: Optional[Exception],
    ) -> None:
        """
        Bedrock API 호출 span을 기록함

        Args:
            operation_name: Bedrock API 작업 이름 (예: InvokeModel, InvokeModelWithResponseStream)
            api_params: API 파라미터
            response: API 응답
            duration: 소요 시간 (밀리초)
            status: 상태 (OK 또는 ERROR)
            trace_id: Trace ID
            parent_span_id: Parent Span ID
            error: 에러 (있는 경우)
        """
        span_id = generate_span_id()
        timestamp = datetime.utcnow().isoformat() + "Z"

        # 모델 ID 추출
        model_id = api_params.get("modelId", "unknown")

        # 토큰 정보 추출 (응답이 있고 status가 OK인 경우)
        input_tokens = None
        output_tokens = None

        if response and status == "OK":
            # Bedrock 응답에서 토큰 정보 추출
            try:
                # InvokeModel 응답의 경우
                if "body" in response:
                    # body를 읽기 전에 원본을 보존하기 위해 내용을 복사
                    original_body = response["body"]
                    body_bytes = original_body.read()

                    # 읽은 내용으로 JSON 파싱하여 토큰 정보 추출
                    # body_bytes는 bytes이므로 decode 필요
                    if isinstance(body_bytes, bytes):
                        body_str = body_bytes.decode("utf-8")
                    else:
                        body_str = body_bytes
                    response_body = json.loads(body_str)

                    # Claude 모델의 경우
                    if "usage" in response_body:
                        input_tokens = response_body["usage"].get("input_tokens")
                        output_tokens = response_body["usage"].get("output_tokens")
                    # 다른 모델의 경우 여기에 추가

                    # body를 읽은 후 원본을 복원하여 bedrock_service에서 사용할 수 있도록 함
                    # io.BytesIO를 사용하여 새로운 스트림 생성
                    from botocore.response import StreamingBody

                    response["body"] = StreamingBody(
                        io.BytesIO(body_bytes),
                        (
                            original_body._content_length
                            if hasattr(original_body, "_content_length")
                            else len(body_bytes)
                        ),
                    )
            except Exception:
                # 토큰 정보 추출 실패 시 무시
                pass

        span: SpanEntry = {
            "type": "span",
            "timestamp": timestamp,
            "service_name": self.service_name,
            "environment": self.environment,
            "trace_id": trace_id,
            "span_id": span_id,
            "parent_span_id": parent_span_id,
            "name": f"Bedrock {operation_name}",
            "kind": "CLIENT",
            "duration_ms": duration,
            "status": status,  # type: ignore
            "bedrock_model_id": model_id,
            "bedrock_operation": operation_name,
            "bedrock_input_tokens": input_tokens,
            "bedrock_output_tokens": output_tokens,
        }

        self.batch_sender.add(span)



"""
모니터링 SDK 메인 클래스
FastAPI 애플리케이션에 간편한 통합을 제공함

역할: SDK의 진입점으로, 모든 collector를 초기화하고 설정을 관리함
사용: main.py에서 MonitoringSDK.init(app, config) 호출로 전체 SDK 활성화
"""

import logging
import atexit
import asyncio
from typing import Optional
from .types.interfaces import MonitoringConfig
from .transport.batch_sender import BatchSender
from .tracing.http_tracer import HttpTracerMiddleware
from .collectors.log_collector import LogCollectorHandler
from .collectors.http_client_collector import HttpClientCollector
from .collectors.bedrock_collector import BedrockCollector


class MonitoringSDK:
    """
    모니터링 SDK 메인 클래스
    """

    _instance: Optional['MonitoringSDK'] = None

    def __init__(self, config: MonitoringConfig):
        """
        MonitoringSDK 초기화 (내부용)

        Args:
            config: 모니터링 설정
        """
        self.config = self._validate_and_normalize_config(config)
        self.batch_sender = BatchSender(self.config)

    @classmethod
    def init(cls, app, config: MonitoringConfig) -> 'MonitoringSDK':
        """
        모니터링 SDK를 초기화함

        역할:
        1. SDK 인스턴스를 생성함 (싱글톤)
        2. BatchSender 시작 (배치 전송 타이머 시작)
        3. 모든 collector 자동 등록 (HTTP, Log, HTTP Client, Bedrock)
        4. Shutdown hook 설정

        사용: main.py에서 단 한 번만 호출함

        Args:
            app: FastAPI 애플리케이션 인스턴스
            config: 모니터링 설정

        Returns:
            MonitoringSDK 인스턴스
        """
        if cls._instance:
            print('[MonitoringSDK] SDK already initialized. Using existing instance.')
            return cls._instance

        sdk = cls(config)
        cls._instance = sdk

        sdk._setup_monitoring(app)

        print(f"[MonitoringSDK] Initialized for service: {config['service_name']}")
        print(f"[MonitoringSDK] Environment: {config.get('environment', 'development')}")
        print(f"[MonitoringSDK] Log Endpoint: {sdk.config.get('log_endpoint')}")
        print(f"[MonitoringSDK] Trace Endpoint: {sdk.config.get('trace_endpoint')}")

        return sdk

    def _setup_monitoring(self, app) -> None:
        """
        모든 모니터링 기능을 설정함

        역할: 설정에 따라 각 collector를 활성화함

        Args:
            app: FastAPI 애플리케이션
        """
        service_name = self.config['service_name']
        environment = self.config.get('environment', 'development')

        if self.config.get('enable_http_tracking', True):
            self._setup_http_tracing(app, service_name, environment)

        if self.config.get('enable_log_tracking', True):
            self._setup_log_collection(service_name, environment)

        if self.config.get('enable_http_client_tracking', True):
            self._setup_http_client_tracking(service_name, environment)

        if self.config.get('enable_bedrock_tracking', True):
            self._setup_bedrock_tracking(service_name, environment)

        self._setup_shutdown_hooks()

    def _setup_http_tracing(self, app, service_name: str, environment: str) -> None:
        """
        HTTP 요청 추적을 설정함

        역할: HttpTracerMiddleware를 FastAPI에 추가하여 모든 HTTP 요청을 추적함

        Args:
            app: FastAPI 애플리케이션
            service_name: 서비스 이름
            environment: 환경 설정
        """
        middleware = HttpTracerMiddleware(
            app=app,
            batch_sender=self.batch_sender,
            service_name=service_name,
            environment=environment
        )

        app.add_middleware(HttpTracerMiddleware.__bases__[0], dispatch=middleware.dispatch)
        print('[MonitoringSDK] HTTP tracing enabled')

    def _setup_log_collection(self, service_name: str, environment: str) -> None:
        """
        로그 수집을 설정함 (Python logging handler)

        역할: Python root logger에 LogCollectorHandler를 추가함

        Args:
            service_name: 서비스 이름
            environment: 환경 설정
        """
        try:
            handler = LogCollectorHandler(
                batch_sender=self.batch_sender,
                service_name=service_name,
                environment=environment
            )

            # Root logger에 handler 추가
            root_logger = logging.getLogger()
            root_logger.addHandler(handler)

            print('[MonitoringSDK] Log collection enabled')
        except Exception as e:
            print(f'[MonitoringSDK] Failed to setup log collection: {e}')

    def _setup_http_client_tracking(self, service_name: str, environment: str) -> None:
        """
        HTTP 클라이언트 추적을 설정함 (httpx, requests)

        역할: httpx와 requests에 instrumentation을 설정하여 외부 API 호출을 추적함

        Args:
            service_name: 서비스 이름
            environment: 환경 설정
        """
        collector = HttpClientCollector(
            batch_sender=self.batch_sender,
            service_name=service_name,
            environment=environment
        )

        collector.setup_httpx_instrumentation()
        collector.setup_requests_instrumentation()

        print('[MonitoringSDK] HTTP client tracking enabled')

    def _setup_bedrock_tracking(self, service_name: str, environment: str) -> None:
        """
        Bedrock API 호출 추적을 설정함 (boto3)

        역할: boto3 bedrock-runtime에 instrumentation을 설정하여 Bedrock 호출을 추적함

        Args:
            service_name: 서비스 이름
            environment: 환경 설정
        """
        collector = BedrockCollector(
            batch_sender=self.batch_sender,
            service_name=service_name,
            environment=environment
        )

        collector.setup_bedrock_instrumentation()

        print('[MonitoringSDK] Bedrock tracking enabled')

    def _setup_shutdown_hooks(self) -> None:
        """
        우아한 종료 hook을 설정함

        역할: 앱 종료 시 버퍼에 남은 데이터를 모두 전송함
        """
        def shutdown_handler():
            print('[MonitoringSDK] Shutting down, flushing data...')
            # asyncio 이벤트 루프에서 shutdown 실행
            try:
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    loop.create_task(self.batch_sender.shutdown())
                else:
                    loop.run_until_complete(self.batch_sender.shutdown())
            except Exception as e:
                print(f'[MonitoringSDK] Error during shutdown: {e}')
            print('[MonitoringSDK] Shutdown complete')

        atexit.register(shutdown_handler)

    def _validate_and_normalize_config(self, config: MonitoringConfig) -> MonitoringConfig:
        """
        설정을 검증하고 정규화함

        역할: 필수 값 검증 및 기본값 적용

        Args:
            config: 사용자 제공 설정

        Returns:
            검증되고 기본값이 적용된 설정

        Raises:
            ValueError: 필수 설정이 없는 경우
        """
        if not config.get('api_key'):
            raise ValueError('[MonitoringSDK] api_key is required')

        if not config.get('service_name'):
            raise ValueError('[MonitoringSDK] service_name is required')

        # endpoint가 제공되면 log_endpoint와 trace_endpoint의 기본값으로 사용
        # 개별 엔드포인트가 명시되면 그것을 우선 사용
        log_endpoint = config.get('log_endpoint')
        trace_endpoint = config.get('trace_endpoint')

        if config.get('endpoint'):
            # 하위 호환성: 기존 endpoint가 제공되면 /logs, /traces 경로를 추가
            log_endpoint = log_endpoint or f"{config['endpoint']}/sdk/logs"
            trace_endpoint = trace_endpoint or f"{config['endpoint']}/sdk/traces"

        if not log_endpoint:
            raise ValueError('[MonitoringSDK] log_endpoint is required (or provide endpoint)')

        if not trace_endpoint:
            raise ValueError('[MonitoringSDK] trace_endpoint is required (or provide endpoint)')

        # 정규화된 설정 반환
        normalized_config: MonitoringConfig = {
            **config,
            'log_endpoint': log_endpoint,
            'trace_endpoint': trace_endpoint,
            'environment': config.get('environment', 'development'),
            'batch_size': config.get('batch_size', 100),
            'flush_interval': config.get('flush_interval', 5),
            'enable_log_tracking': config.get('enable_log_tracking', True),
            'enable_http_tracking': config.get('enable_http_tracking', True),
            'enable_http_client_tracking': config.get('enable_http_client_tracking', True),
            'enable_bedrock_tracking': config.get('enable_bedrock_tracking', True)
        }

        return normalized_config

    async def flush(self) -> None:
        """
        데이터를 수동으로 flush함

        역할: 테스트나 특수한 경우 즉시 데이터를 전송함
        사용: await sdk.flush()
        """
        await self.batch_sender.flush()

    def get_buffer_size(self) -> int:
        """
        현재 버퍼 크기를 조회함

        역할: 모니터링이나 디버깅 목적으로 버퍼 상태 확인

        Returns:
            버퍼에 쌓인 데이터 개수
        """
        return self.batch_sender.get_buffer_size()



"""
모니터링 데이터를 Producer 서버로 전송하는 배치 발송자
배치 전송 전략을 구현함: 버퍼가 가득 차거나 일정 시간이 경과하면 전송함

역할: 로그와 Span 데이터를 별도의 메모리 버퍼에 모아서 각각의 엔드포인트로
      일괄 전송하여 네트워크 오버헤드를 최소화함
"""

import asyncio
import threading
from typing import List
import httpx
from ..types.interfaces import DataEntry, LogEntry, SpanEntry, MonitoringConfig


class BatchSender:
    """
    모니터링 데이터를 배치로 전송하는 클래스
    """

    def __init__(self, config: MonitoringConfig):
        """
        BatchSender 초기화

        Args:
            config: 모니터링 설정
        """
        self.config = config
        self.log_buffer: List[LogEntry] = []
        self.span_buffer: List[SpanEntry] = []
        self._lock = threading.Lock()
        self._is_flushing = False
        self._flush_task: Optional[asyncio.Task] = None
        self._should_stop = False

        # HTTP 클라이언트 설정
        self.log_client = httpx.AsyncClient(
            base_url=config.get('log_endpoint', ''),
            timeout=10.0,
            headers={
                'Content-Type': 'application/json',
                'X-API-Key': config['api_key']
            }
        )

        self.trace_client = httpx.AsyncClient(
            base_url=config.get('trace_endpoint', ''),
            timeout=10.0,
            headers={
                'Content-Type': 'application/json',
                'X-API-Key': config['api_key']
            }
        )

        # 주기적 플러시 타이머 시작
        self._start_flush_timer()

    def add(self, entry: DataEntry) -> None:
        """
        버퍼에 데이터 엔트리를 추가함

        역할: 새 데이터를 타입에 따라 적절한 버퍼에 추가하고, 버퍼가 가득 차면 자동으로 전송함
        사용: LogCollectorHandler, HttpTracerMiddleware, HttpClientCollector,
              BedrockCollector에서 데이터를 수집할 때마다 호출됨

        Args:
            entry: Log 또는 Span 엔트리
        """
        with self._lock:
            # 타입에 따라 적절한 버퍼에 추가
            if entry['type'] == 'log':
                self.log_buffer.append(entry)  # type: ignore
            elif entry['type'] == 'span':
                self.span_buffer.append(entry)  # type: ignore

            batch_size = self.config.get('batch_size', 100)

            # 각 버퍼가 배치 크기에 도달하면 플러시
            if len(self.log_buffer) >= batch_size or len(self.span_buffer) >= batch_size:
                # 백그라운드에서 플러시 실행
                asyncio.create_task(self._flush_async())

    async def flush(self) -> None:
        """
        버퍼의 데이터를 Producer 서버로 전송함

        역할: 로그와 트레이스 버퍼에 쌓인 모든 데이터를 각각의 엔드포인트로 전송함
        사용: 버퍼가 가득 차거나(batch_size), 일정 시간마다(flush_interval), 또는 앱 종료 시 호출됨
        """
        await self._flush_async()

    async def _flush_async(self) -> None:
        """
        비동기로 버퍼를 플러시함 (내부 메서드)
        """
        if self._is_flushing:
            return

        with self._lock:
            # 두 버퍼 모두 비어있으면 리턴
            if len(self.log_buffer) == 0 and len(self.span_buffer) == 0:
                return

            self._is_flushing = True

            # 버퍼 복사 및 초기화
            logs_to_send = self.log_buffer.copy()
            spans_to_send = self.span_buffer.copy()
            self.log_buffer.clear()
            self.span_buffer.clear()

        try:
            # 로그와 트레이스를 병렬로 전송
            tasks = []

            if logs_to_send:
                tasks.append(self._send_logs(logs_to_send))

            if spans_to_send:
                tasks.append(self._send_traces(spans_to_send))

            await asyncio.gather(*tasks, return_exceptions=True)

        except Exception as e:
            print(f'[MonitoringSDK] Failed to send data to Producer: {e}')
        finally:
            self._is_flushing = False

    async def _send_logs(self, logs: List[LogEntry]) -> None:
        """
        로그 데이터를 전송함

        Args:
            logs: 전송할 로그 목록
        """
        try:
            response = await self.log_client.post('', json=logs)
            response.raise_for_status()
        except Exception as e:
            print(f'[MonitoringSDK] Failed to send logs: {e}')

    async def _send_traces(self, spans: List[SpanEntry]) -> None:
        """
        트레이스 데이터를 전송함

        Args:
            spans: 전송할 span 목록
        """
        try:
            response = await self.trace_client.post('', json=spans)
            response.raise_for_status()
        except Exception as e:
            print(f'[MonitoringSDK] Failed to send traces: {e}')

    def _start_flush_timer(self) -> None:
        """
        주기적인 플러시 타이머를 시작함

        역할: 일정 시간(기본 5초)마다 자동으로 버퍼를 비우도록 타이머를 설정함
        사용: BatchSender 생성 시 자동으로 호출되어 주기적 전송을 활성화함
        """
        async def periodic_flush():
            interval = self.config.get('flush_interval', 5)
            while not self._should_stop:
                await asyncio.sleep(interval)
                if not self._should_stop:
                    await self._flush_async()

        # 백그라운드 태스크로 실행
        loop = asyncio.get_event_loop()
        self._flush_task = loop.create_task(periodic_flush())

    async def shutdown(self) -> None:
        """
        플러시 타이머를 중지하고 남은 데이터를 전송함

        역할: 앱 종료 시 타이머를 정리하고 버퍼에 남은 모든 데이터를 전송함
        사용: MonitoringSDK의 shutdown hook에서 호출되어 데이터 손실을 방지함
        """
        self._should_stop = True

        if self._flush_task:
            self._flush_task.cancel()
            try:
                await self._flush_task
            except asyncio.CancelledError:
                pass

        # 남은 데이터 플러시
        await self._flush_async()

        # HTTP 클라이언트 종료
        await self.log_client.aclose()
        await self.trace_client.aclose()

    def get_buffer_size(self) -> int:
        """
        현재 버퍼 크기를 조회함

        역할: 현재 버퍼에 쌓인 데이터 개수를 반환함 (로그 + 트레이스)
        사용: 모니터링이나 디버깅 목적으로 사용됨

        Returns:
            버퍼에 쌓인 데이터 개수
        """
        with self._lock:
            return len(self.log_buffer) + len(self.span_buffer)

    def get_log_buffer_size(self) -> int:
        """로그 버퍼 크기를 조회함"""
        with self._lock:
            return len(self.log_buffer)

    def get_trace_buffer_size(self) -> int:
        """트레이스 버퍼 크기를 조회함"""
        with self._lock:
            return len(self.span_buffer)



"""
HTTP 클라이언트 수집기
httpx와 requests 요청을 가로채서 외부 API 호출을 추적함

역할: httpx와 requests를 통한 모든 외부 API 호출을 자동으로 추적함
사용: MonitoringSDK.init()에서 자동 등록됨
"""

import time
from datetime import datetime
from typing import Optional
from ..types.interfaces import SpanEntry
from ..transport.batch_sender import BatchSender
from ..tracing.trace_context import get_trace_context
from ..tracing.id_generator import generate_span_id


class HttpClientCollector:
    """
    HTTP 클라이언트 호출을 추적하는 수집기
    """

    def __init__(
        self,
        batch_sender: BatchSender,
        service_name: str,
        environment: str
    ):
        """
        HttpClientCollector 초기화

        Args:
            batch_sender: 배치 전송자
            service_name: 서비스 이름
            environment: 환경 설정
        """
        self.batch_sender = batch_sender
        self.service_name = service_name
        self.environment = environment

    def setup_httpx_instrumentation(self) -> None:
        """
        httpx 라이브러리에 instrumentation을 설정함

        역할: httpx.AsyncClient 사용 시 자동으로 추적하도록 설정
        """
        try:
            import httpx

            # httpx event hooks를 사용하여 추적
            original_init = httpx.AsyncClient.__init__

            def patched_init(client_self, *args, **kwargs):
                # 기존 event_hooks 가져오기
                event_hooks = kwargs.get('event_hooks', {})

                # 요청/응답 훅 추가
                request_hooks = event_hooks.get('request', [])
                response_hooks = event_hooks.get('response', [])

                request_hooks.append(self._httpx_request_hook)
                response_hooks.append(self._httpx_response_hook)

                event_hooks['request'] = request_hooks
                event_hooks['response'] = response_hooks
                kwargs['event_hooks'] = event_hooks

                original_init(client_self, *args, **kwargs)

            httpx.AsyncClient.__init__ = patched_init

        except ImportError:
            # httpx가 설치되지 않은 경우 무시
            pass

    async def _httpx_request_hook(self, request) -> None:
        """httpx 요청 시작 훅"""
        request._monitoring_start_time = time.time()

    async def _httpx_response_hook(self, response) -> None:
        """httpx 응답 훅"""
        trace_context = get_trace_context()
        if not trace_context:
            return

        request = response.request
        if not hasattr(request, '_monitoring_start_time'):
            return

        duration = (time.time() - request._monitoring_start_time) * 1000
        status = 'OK' if response.status_code < 400 else 'ERROR'

        self._record_http_client_span(
            str(request.method),
            str(request.url),
            response.status_code,
            duration,
            status,
            trace_context.trace_id,
            trace_context.span_id
        )

    def setup_requests_instrumentation(self) -> None:
        """
        requests 라이브러리에 instrumentation을 설정함

        역할: requests.request 사용 시 자동으로 추적하도록 설정
        """
        try:
            import requests

            original_request = requests.request

            def patched_request(method, url, **kwargs):
                trace_context = get_trace_context()

                if not trace_context:
                    return original_request(method, url, **kwargs)

                start_time = time.time()
                status = 'ERROR'
                status_code = 0

                try:
                    response = original_request(method, url, **kwargs)
                    status_code = response.status_code
                    status = 'OK' if status_code < 400 else 'ERROR'
                    return response
                except Exception as e:
                    status = 'ERROR'
                    raise
                finally:
                    duration = (time.time() - start_time) * 1000
                    self._record_http_client_span(
                        method.upper(),
                        url,
                        status_code,
                        duration,
                        status,
                        trace_context.trace_id,
                        trace_context.span_id
                    )

            requests.request = patched_request

        except ImportError:
            # requests가 설치되지 않은 경우 무시
            pass

    def _record_http_client_span(
        self,
        method: str,
        url: str,
        status_code: int,
        duration: float,
        status: str,
        trace_id: str,
        parent_span_id: str
    ) -> None:
        """
        HTTP 클라이언트 span을 기록함

        Args:
            method: HTTP 메서드
            url: 요청 URL
            status_code: HTTP 상태 코드
            duration: 소요 시간 (밀리초)
            status: 상태 (OK 또는 ERROR)
            trace_id: Trace ID
            parent_span_id: Parent Span ID
        """
        span_id = generate_span_id()
        timestamp = datetime.utcnow().isoformat() + 'Z'

        span: SpanEntry = {
            'type': 'span',
            'timestamp': timestamp,
            'service_name': self.service_name,
            'environment': self.environment,
            'trace_id': trace_id,
            'span_id': span_id,
            'parent_span_id': parent_span_id,
            'name': f'{method} {url}',
            'kind': 'CLIENT',
            'duration_ms': duration,
            'status': status,  # type: ignore
            'http_method': method,
            'http_url': url,
            'http_status_code': status_code if status_code > 0 else None
        }

        self.batch_sender.add(span)



"""
contextvars를 사용한 Trace Context 관리자
FastAPI의 비동기 작업들 간에 trace context를 관리함

역할: HTTP 요청부터 API 호출, Bedrock 호출까지 동일한 trace_id를 유지하여
      분산 추적을 가능하게 함
"""

from contextvars import ContextVar
from typing import Optional
from ..types.interfaces import TraceContext
from .id_generator import generate_trace_id, generate_span_id

# ContextVar를 사용하여 비동기 컨텍스트에서 trace context 관리
_trace_context: ContextVar[Optional[TraceContext]] = ContextVar('trace_context', default=None)


def get_trace_context() -> Optional[TraceContext]:
    """
    현재 trace context를 조회함

    역할: 현재 실행 중인 비동기 작업의 trace context를 가져옴
    사용: API 호출이나 Bedrock 호출 시 parent span을 찾기 위해 사용됨

    Returns:
        현재 trace context 또는 추적 중이 아니면 None
    """
    return _trace_context.get()


def set_trace_context(context: Optional[TraceContext]) -> None:
    """
    trace context를 설정함

    역할: 특정 trace context를 현재 실행 컨텍스트에 설정함
    사용: HTTP 요청 처리 시작 시 root context를 설정함

    Args:
        context: 설정할 Trace context
    """
    _trace_context.set(context)


def create_root_context() -> TraceContext:
    """
    새로운 root trace context를 생성함 (HTTP 요청용)

    역할: HTTP 요청이 시작될 때 새로운 추적 세션을 시작함
    사용: HTTP Middleware에서 각 HTTP 요청마다 호출됨

    Returns:
        생성된 trace_id와 span_id를 가진 새 trace context
    """
    return TraceContext(
        trace_id=generate_trace_id(),
        span_id=generate_span_id(),
        parent_span_id=None
    )


def create_child_context() -> Optional[TraceContext]:
    """
    현재 context로부터 child span context를 생성함

    역할: API 호출이나 Bedrock 호출 시 child span을 생성함
    사용: HttpClientCollector, BedrockCollector에서 호출되어
          parent span(HTTP 요청)과 연결된 child span을 만듦

    Returns:
        같은 trace_id를 가지지만 새 span_id를 가진 trace context
        또는 현재 context가 없으면 None
    """
    current_context = get_trace_context()

    if not current_context:
        return None

    return TraceContext(
        trace_id=current_context.trace_id,
        span_id=generate_span_id(),
        parent_span_id=current_context.span_id
    )


def get_current_trace_id() -> Optional[str]:
    """
    현재 trace ID를 조회함

    역할: 현재 실행 중인 요청의 trace ID를 반환함
    사용: 로깅 시 trace ID를 함께 기록하고 싶을 때 사용됨

    Returns:
        현재 trace ID 또는 None
    """
    context = get_trace_context()
    return context.trace_id if context else None


def get_current_span_id() -> Optional[str]:
    """
    현재 span ID를 조회함

    역할: 현재 실행 중인 작업의 span ID를 반환함

    Returns:
        현재 span ID 또는 None
    """
    context = get_trace_context()
    return context.span_id if context else None


def get_current_parent_span_id() -> Optional[str]:
    """
    현재 parent span ID를 조회함

    역할: 현재 span의 부모 span ID를 반환함

    Returns:
        현재 parent span ID 또는 None
    """
    context = get_trace_context()
    return context.parent_span_id if context else None


def is_in_traced_context() -> bool:
    """
    현재 추적 context 내에 있는지 확인함

    역할: 현재 HTTP 요청 처리 중인지 확인함

    Returns:
        추적 중이면 True, 아니면 False
    """
    return get_trace_context() is not None



"""
OpenTelemetry 호환 ID 생성 모듈
- trace_id: 32자 hex 문자열 (16 bytes 랜덤)
- span_id: 16자 hex 문자열 (8 bytes 랜덤)
"""

import os
import re


def generate_trace_id() -> str:
    """
    Trace ID를 생성함 (32자 hex 문자열)

    역할: 전체 요청 추적을 식별하는 고유 ID를 생성함
    사용: HTTP 요청마다 새로운 trace_id가 생성되며, 해당 요청 내의 모든 span이 이 ID를 공유함

    Returns:
        32자 16진수 문자열
    """
    return os.urandom(16).hex()


def generate_span_id() -> str:
    """
    Span ID를 생성함 (16자 hex 문자열)

    역할: 개별 작업(HTTP 요청, API 호출, Bedrock 호출 등)을 식별하는 고유 ID를 생성함
    사용: 각 span(작업 단위)마다 새로운 span_id가 생성됨

    Returns:
        16자 16진수 문자열
    """
    return os.urandom(8).hex()


def is_valid_trace_id(trace_id: str) -> bool:
    """
    문자열이 유효한 Trace ID인지 검증함

    역할: Trace ID가 OpenTelemetry 표준 형식(32자 hex)에 맞는지 확인함

    Args:
        trace_id: 검증할 Trace ID

    Returns:
        유효하면 True, 아니면 False
    """
    return bool(re.match(r'^[0-9a-f]{32}$', trace_id))


def is_valid_span_id(span_id: str) -> bool:
    """
    문자열이 유효한 Span ID인지 검증함

    역할: Span ID가 OpenTelemetry 표준 형식(16자 hex)에 맞는지 확인함

    Args:
        span_id: 검증할 Span ID

    Returns:
        유효하면 True, 아니면 False
    """
    return bool(re.match(r'^[0-9a-f]{16}$', span_id))



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="panopticon-monitoring",
    version="0.1.2",
    author="Woongno",
    description="FastAPI monitoring SDK for logs and distributed tracing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/panopticon-jungle10/mornitoring_sdk",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Framework :: FastAPI",
    ],
    python_requires=">=3.8",
    install_requires=[
        "httpx>=0.24.0",
        "starlette>=0.27.0",
    ],
    extras_require={
        "requests": ["requests>=2.28.0"],
        "boto3": ["boto3>=1.26.0"],
        "all": [
            "requests>=2.28.0",
            "boto3>=1.26.0",
        ],
    },
    keywords=[
        "fastapi",
        "monitoring",
        "tracing",
        "logging",
        "observability",
        "opentelemetry",
        "bedrock",
        "aws",
    ],
    project_urls={
        "Bug Reports": "https://github.com/panopticon-jungle10/mornitoring_sdk/issues",
        "Source": "https://github.com/panopticon-jungle10/mornitoring_sdk",
    },
)



"""
로그 수집을 위한 Python logging Handler
모든 로그 레벨을 자동으로 캡처하여 Producer로 전송함

역할: Python logger에 자동으로 추가되어 모든 로그를 수집함
사용: MonitoringSDK.init()에서 logger에 자동 등록됨
"""

import logging
from datetime import datetime
from typing import Optional
from ..types.interfaces import LogEntry
from ..transport.batch_sender import BatchSender
from ..tracing.trace_context import get_current_trace_id


class LogCollectorHandler(logging.Handler):
    """
    로그 수집을 위한 logging.Handler
    """

    def __init__(
        self,
        batch_sender: BatchSender,
        service_name: str,
        environment: str
    ):
        """
        LogCollectorHandler 초기화

        Args:
            batch_sender: 배치 전송자
            service_name: 서비스 이름
            environment: 환경 설정
        """
        super().__init__()
        self.batch_sender = batch_sender
        self.service_name = service_name
        self.environment = environment

    def emit(self, record: logging.LogRecord) -> None:
        """
        Python에서 로그가 발생할 때마다 호출됨

        역할:
        1. Python에서 전달된 로그 정보를 받음
        2. LogEntry 형식으로 변환함
        3. BatchSender에 추가함

        사용: logger.log(), logger.error(), logger.warning() 등 모든 로그에서 자동 호출됨

        Args:
            record: logging.LogRecord 객체
        """
        # httpx, urllib3 로그 무시 (무한 루프 방지)
        if record.name.startswith(('httpx', 'urllib3', 'httpcore')):
            return

        try:
            # 현재 trace context에서 trace_id 가져오기
            trace_id = get_current_trace_id()

            log_entry: LogEntry = {
                'type': 'log',
                'timestamp': datetime.utcnow().isoformat() + 'Z',
                'service_name': self.service_name,
                'environment': self.environment,
                'level': record.levelname.lower(),
                'message': self.format(record),
                'context': record.name,
                'trace': self._get_trace_info(record),
                'trace_id': trace_id  # 트레이스 도중이면 trace_id, 아니면 None
            }

            self.batch_sender.add(log_entry)
        except Exception:
            # 로그 수집 실패가 앱에 영향을 주지 않도록 무시
            self.handleError(record)

    def _get_trace_info(self, record: logging.LogRecord) -> Optional[str]:
        """
        로그 레코드에서 트레이스 정보를 추출함

        Args:
            record: logging.LogRecord 객체

        Returns:
            트레이스 정보 문자열 또는 None
        """
        if hasattr(record, 'exc_info') and record.exc_info:
            return self.formatter.formatException(record.exc_info) if self.formatter else None
        if hasattr(record, 'stack_info') and record.stack_info:
            return record.stack_info
        return None



"""
HTTP 추적 미들웨어
모든 HTTP 요청을 자동으로 추적하고 root span을 생성함

역할: FastAPI의 모든 HTTP 요청을 가로채서 추적 시작점(root span)을 만듦
사용: MonitoringSDK.init()에서 자동으로 등록되어 모든 라우터 요청을 추적함
"""

import time
from datetime import datetime
from typing import Callable, Optional
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import Response
from ..types.interfaces import SpanEntry
from ..transport.batch_sender import BatchSender
from .trace_context import create_root_context, set_trace_context


class HttpTracerMiddleware(BaseHTTPMiddleware):
    """
    HTTP 요청을 추적하는 미들웨어
    """

    def __init__(
        self, app, batch_sender: BatchSender, service_name: str, environment: str
    ):
        """
        HttpTracerMiddleware 초기화

        Args:
            app: FastAPI 애플리케이션
            batch_sender: 배치 전송자
            service_name: 서비스 이름
            environment: 환경 설정
        """
        super().__init__(app)
        self.batch_sender = batch_sender
        self.service_name = service_name
        self.environment = environment

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        """
        HTTP 요청을 가로채서 추적함

        역할:
        1. 요청 시작 시 root trace context 생성 (trace_id, span_id)
        2. set_trace_context로 context 저장 (ContextVar에 저장됨)
        3. 요청 처리
        4. 응답 시 duration 계산하고 Root SpanEntry 생성
        5. BatchSender에 전송

        Args:
            request: Starlette Request
            call_next: 다음 핸들러

        Returns:
            Response
        """
        # 헬스체크 경로는 trace를 생성하지 않음
        if request.url.path == "/health" or request.url.path == "/":
            return await call_next(request)

        # Root trace context 생성 - 전체 추적의 시작점
        trace_context = create_root_context()
        start_time = time.time()
        start_timestamp = datetime.utcnow().isoformat() + "Z"

        # Trace context를 ContextVar에 저장
        # 이제 이 요청 처리 중 발생하는 모든 API 호출, Bedrock 호출이 같은 trace_id를 공유함
        set_trace_context(trace_context)

        status_code = 500
        status = "ERROR"

        try:
            response = await call_next(request)
            status_code = response.status_code
            status = "OK" if status_code < 400 else "ERROR"
            return response
        except Exception as e:
            status = "ERROR"
            raise
        finally:
            duration = (time.time() - start_time) * 1000  # Convert to milliseconds

            # HTTP Span 기록
            self._record_span(
                trace_context.trace_id,
                trace_context.span_id,
                None,
                request,
                status_code,
                start_timestamp,
                duration,
                status,
            )

    def _record_span(
        self,
        trace_id: str,
        span_id: str,
        parent_span_id: Optional[str],
        request: Request,
        status_code: int,
        timestamp: str,
        duration: float,
        status: str,
    ) -> None:
        """
        HTTP Span을 기록함

        Args:
            trace_id: Trace ID
            span_id: Span ID
            parent_span_id: Parent Span ID (root는 None)
            request: Starlette Request 객체
            status_code: HTTP 상태 코드
            timestamp: 시작 시각
            duration: 소요 시간 (밀리초)
            status: 상태 (OK 또는 ERROR)
        """
        method = request.method
        path = request.url.path
        url = str(request.url)

        span: SpanEntry = {
            "type": "span",
            "timestamp": timestamp,
            "service_name": self.service_name,
            "environment": self.environment,
            "trace_id": trace_id,
            "span_id": span_id,
            "parent_span_id": parent_span_id,
            "name": f"{method} {path}",
            "kind": "SERVER",
            "duration_ms": duration,
            "status": status,  # type: ignore
            "http_method": method,
            "http_path": path,
            "http_url": url,
            "http_status_code": status_code,
        }

        self.batch_sender.add(span)



"""
@panopticon/fastapi-monitoring-sdk

간편한 FastAPI 모니터링 SDK (로그 및 분산 추적)

사용법:
```python
from fastapi import FastAPI
from panopticon_monitoring import MonitoringSDK

app = FastAPI()

MonitoringSDK.init(app, {
    'api_key': 'your-api-key',
    'endpoint': 'https://producer.woongno-monitoring.com',
    'service_name': 'llm-service',
    'environment': 'production'
})
```
"""

# SDK 메인 클래스 export
from .monitoring_sdk import MonitoringSDK

# 타입 정의 export
from .types.interfaces import (
    MonitoringConfig,
    LogEntry,
    SpanEntry,
    DataEntry,
    TraceContext,
    SpanOptions,
    ActiveSpan
)

# ID 생성 함수 export
from .tracing.id_generator import (
    generate_trace_id,
    generate_span_id,
    is_valid_trace_id,
    is_valid_span_id
)

# Trace Context 관리 함수 export
from .tracing.trace_context import (
    get_trace_context,
    set_trace_context,
    create_root_context,
    create_child_context,
    get_current_trace_id,
    get_current_span_id,
    get_current_parent_span_id,
    is_in_traced_context
)

# 고급 사용을 위한 collector 클래스 export
from .transport.batch_sender import BatchSender
from .tracing.http_tracer import HttpTracerMiddleware
from .collectors.log_collector import LogCollectorHandler
from .collectors.http_client_collector import HttpClientCollector
from .collectors.bedrock_collector import BedrockCollector

__version__ = '0.1.0'

__all__ = [
    # 메인 클래스
    'MonitoringSDK',

    # 타입
    'MonitoringConfig',
    'LogEntry',
    'SpanEntry',
    'DataEntry',
    'TraceContext',
    'SpanOptions',
    'ActiveSpan',

    # ID 생성
    'generate_trace_id',
    'generate_span_id',
    'is_valid_trace_id',
    'is_valid_span_id',

    # Trace Context
    'get_trace_context',
    'set_trace_context',
    'create_root_context',
    'create_child_context',
    'get_current_trace_id',
    'get_current_span_id',
    'get_current_parent_span_id',
    'is_in_traced_context',

    # Collectors
    'BatchSender',
    'HttpTracerMiddleware',
    'LogCollectorHandler',
    'HttpClientCollector',
    'BedrockCollector',
]
